/*
 *  PS3 flash memory utility.
 *
 *  Copyright (C) 2006 Sony Computer Entertainment Inc.
 *  Copyright 2006 Sony Corp.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#if defined(HAVE_CONFIG_H)
#include "config.h"
#endif
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <inttypes.h>
#include <ps3-flash.h>

#define PS3_UTILS_APP_NAME "ps3-flash-util"

#if defined(PACKAGE_VERSION) && defined(PACKAGE_NAME)
#define PS3_UTILS_VERSION PS3_UTILS_APP_NAME " (" PACKAGE_NAME ") " PACKAGE_VERSION "\n\n"
#else
#define PS3_UTILS_VERSION ""
#endif

#if defined(PACKAGE_BUGREPORT)
#define PS3_UTILS_BUGREPORT "Send bug reports to " PACKAGE_BUGREPORT "\n\n"
#else
#define PS3_UTILS_BUGREPORT ""
#endif

#if defined(DEBUG)
#define DBG(_args...) do {printf(_args);} while(0)
#else
static inline int __attribute__ ((format (printf, 1, 2)))
 
	DBG(__attribute__((unused)) const char *fmt, ...) {return 0;}
#endif

static const char flash_dev[] = "/dev/ps3flash";

/**
 * enum db_op_type - Types of database operations.
 */

enum db_op_type {op_undef = 0, op_print, op_remove, op_set_16, op_set_32,
	op_set_64};

/**
 * struct db_op_entry - A database operations list entry.
 */

struct db_op_entry {
	enum db_op_type type;
	struct os_area_db_id id;
	uint64_t value;
	struct db_op_entry *next;
};

static void print_usage(void)
{
	printf(PS3_UTILS_VERSION);
	printf("SYNOPSIS\n");
	printf("     ps3-flash-util [-hsgorzFLv] [-d flash-dev] [-w image-file] [-P owner key]\n");
	printf("                    [-D owner key dword] [-W owner key word]\n");
	printf("                    [-H owner key half] [-R owner key]\n");
	printf("\n");
	printf(PS3_UTILS_BUGREPORT);
}

/**
 * enum opt_value - Tri-state options variables.
 */

enum opt_value {opt_undef = 0, opt_yes, opt_no};

/**
 * struct opts - User specified options.
 */

struct opts {
	unsigned int verbosity;
	unsigned int show_help;
	unsigned int show_settings;
	unsigned int boot_other_os;
	unsigned int compressed_image;
	unsigned int db_format;
	unsigned int db_list_owners;
	struct db_op_entry* db_op_list;
	const char *device;
	const char *image;
};

static void opts_dump(const struct opts* opts)
{
	struct db_op_entry *p;

	DBG("verbosity        %d\n", opts->verbosity);
	DBG("show_help        %d\n", opts->show_help);
	DBG("show_settings    %d\n", opts->show_settings);
	DBG("boot_other_os    %d\n", opts->boot_other_os);
	DBG("compressed_image %d\n", opts->compressed_image);
	DBG("db_format        %d\n", opts->db_format);
	DBG("db_list_owners   %d\n", opts->db_list_owners);
	DBG("device           %s\n", opts->device);
	DBG("image            %s\n", opts->image);

	DBG("db_op_list       %p\n", opts->db_op_list);
	for (p = opts->db_op_list; p; p = p->next) {
		DBG("  %p: type %d, (%d:%d) = %llu\n", p, p->type, p->id.owner,
			p->id.key, (unsigned long long)p->value);
	}
}

static void show_settings(const struct os_area_header *h,
	const struct os_area_params *p, const struct os_area_db *db)
{
	int64_t rtc_diff;

	printf(PS3_UTILS_VERSION);
	os_area_header_dump(h, " header ", 1);
	os_area_params_dump(p, " param  ", 2);

	if (!db)
		printf(" db     :3: not found\n");
	else {
		int result;

		os_area_db_dump_header(db,    " db     ", 3);

		result = os_area_db_get_rtc_diff(db, &rtc_diff);
		if(!result)
			printf(" rtc    :4: %lld\n", (long long int)rtc_diff);
	}
}

static int write_image(FILE *dev, struct os_area_header *h, const char *image)
{
	int result;
	FILE *img;

	if (image[0] == '-' && image[1] == '\0') {
		DBG("%s:%d: read on stdin\n", __func__, __LINE__);
		img = stdin;
	} else
		img = fopen(image, "r");

	if (!img) {
		DBG("%s:%d: fopen failed.\n", __func__, __LINE__);
		perror(image);
		return -1;
	}

	result = os_area_image_write(img, h, dev);

	return result;
}

static void db_list_owners(void)
{
	char s[2048];

	os_area_db_list_owners(s, sizeof(s));
	printf(PS3_UTILS_VERSION);
	printf("%s\n", s);
}

/**
 * pickup_str - Pickup a string option from the command line.
 */

static int pickup_str(char *argv[], const char **pp)
{
	DBG(" -> %s:%d: %s %s\n", __func__, __LINE__, argv[0], argv[1]);

	if (argv[0][2]) {
		fprintf(stderr, "bad option: '%c' in '%s'\n", argv[0][2],
			argv[0]);
		return -1;
	}

	if (!argv[1] || (argv[1][0] == '-' && argv[1][1])) {
		fprintf(stderr, "empty option: -%c in '%s'\n", argv[0][1],
			argv[0]);
		return -1;
	}

	*pp = argv[1];
	return 0;
}

/**
 * pickup_db_op_entry - Pickup a database option from the command line.
 */

static int pickup_db_op_entry(char *argv[], enum db_op_type type,
	unsigned int required, struct db_op_entry** list)
{
	int result;
	struct db_op_entry e;
	struct db_op_entry *p;
	unsigned long long values[3];
	unsigned int i;

	memset(&e, 0, sizeof(struct db_op_entry));
	values[0] = values[1] = values[2] = 0;

	if (argv[0][2]) {
		fprintf(stderr, "bad option: '%c' in '%s'\n", argv[0][2],
			argv[0]);
		return -1;
	}

	for (i = 0; i < required; i++) {
		if (!argv[i + 1]) {
			if (required > i) {
				fprintf(stderr, "missing option: -%c at '%s'\n",
					argv[0][1], argv[i + 1]);
				return -1;
			}
			break;
		}
	
		result = sscanf(argv[i + 1], "%Li", &values[i]);
	
		if (result != 1) {
			fprintf(stderr, "bad option: -%c at '%s'\n",
				argv[0][1], argv[i + 1]);
			return -1;
		}

		//DBG("%s:%d: (%u) found '%s' => %llu (%llxh)\n", __func__, __LINE__,
		//	i, argv[i + 1], values[i], values[i]);
	}

	if((int)values[0] != OS_AREA_DB_OWNER_ANY
		&& (int)values[0] >= OS_AREA_DB_OWNER_MAX) {
		fprintf(stderr, "bad owner value: %u\n",
			(unsigned int)values[0]);
		return -1;
	}

	if((int)values[1] != OS_AREA_DB_KEY_ANY
		&& (int)values[1] >= OS_AREA_DB_KEY_MAX) {
		fprintf(stderr, "bad key value: %u\n",
			(unsigned int)values[1]);
		return -1;
	}

	e.type = type;
	e.id.owner = values[0];
	e.id.key = values[1];
	e.value = values[2];

	DBG("%s:%d: %s (%d:%d) = %llu (%llxh)\n", __func__, __LINE__,
		argv[0], (unsigned int)e.id.owner, (unsigned int)e.id.key,
		(unsigned long long)e.value, (unsigned long long)e.value);

	if (!*list) {
		/* init list */
		*list = malloc(sizeof(struct db_op_entry));
		p = *list;
	} else {
		/* add to list */
		p = *list;
		while(p->next)
			p = p->next;
		p->next = malloc(sizeof(struct db_op_entry));
		p = p->next;
	}

	*p = e;
	return 0;
}

/**
 * opts_parse - Parse the command line options.
 */

static int opts_parse(struct opts* opts, int argc, char *argv[])
{
	int result;
	int got_work = 0;
	int i;

	memset(opts, 0, sizeof(struct opts));

	if (argc < 2)
		return -1;

	for (i = 1; i < argc; i++) {
		const char* p = argv[i];

		if (*p != '-') {
			fprintf(stderr, "unknown option: %s\n\n", p);
			return -1;
		}

next_opt:
		p++;
		switch (*p) {
		case '\0':
			DBG("%s:%d: done\n", __func__, __LINE__);
			break;
		case 'h':
			opts->show_help = opt_yes;
			DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'd':
			result = pickup_str(&argv[i], &opts->device);
			if (result) {
				fprintf(stderr, "%s:%d: parse failed: -%c\n\n",
					__func__, __LINE__, *p);
				return -1;
			}
			i++;
			got_work++;
			DBG("%s:%d: found %c, device '%s'\n", __func__,
				__LINE__, *p, opts->device);
			continue;
		case 's':
			opts->show_settings = opt_yes;
			got_work++;
			DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'w':
			result = pickup_str(&argv[i], &opts->image);
			if (result) {
				fprintf(stderr, "%s:%d: parse failed: -%c\n\n",
					__func__, __LINE__, *p);
				return -1;
			}
			i++;
			got_work++;
			DBG("%s:%d: found %c, image '%s'\n", __func__,
				__LINE__, *p, opts->image);
			continue;
		case 'g':
			opts->boot_other_os = opt_no;
			got_work++;
			DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'o':
			opts->boot_other_os = opt_yes;
			got_work++;
			DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'z':
			opts->compressed_image = opt_yes;
			got_work++;
			DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'r':
			opts->compressed_image = opt_no;
			got_work++;
			DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'v':
			opts->verbosity++;
			DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'F':
			opts->db_format = opt_yes;
			got_work++;
			DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'P':
			result = pickup_db_op_entry(&argv[i], op_print, 2,
				&opts->db_op_list);
			if (result) {
				fprintf(stderr, "%s:%d: parse failed: -%c\n\n",
					__func__, __LINE__, *p);
				return -1;
			}
			i += 2;
			got_work++;
			//DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'D':
			result = pickup_db_op_entry(&argv[i], op_set_64, 3,
				&opts->db_op_list);
			if (result) {
				fprintf(stderr, "%s:%d: parse failed: -%c\n\n",
					__func__, __LINE__, *p);
				return -1;
			}
			i += 3;
			got_work++;
			//DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'W':
			result = pickup_db_op_entry(&argv[i], op_set_32, 3,
				&opts->db_op_list);
			if (result) {
				fprintf(stderr, "parse failed: -%c\n\n", *p);
				return -1;
			}
			i += 3;
			got_work++;
			//DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'H':
			result = pickup_db_op_entry(&argv[i], op_set_16, 3,
				&opts->db_op_list);
			if (result) {
				fprintf(stderr, "%s:%d: parse failed: -%c\n\n",
					__func__, __LINE__, *p);
				return -1;
			}
			i += 3;
			got_work++;
			//DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'R':
			result = pickup_db_op_entry(&argv[i], op_remove, 2,
				&opts->db_op_list);
			if (result) {
				fprintf(stderr, "%s:%d: parse failed: -%c\n\n",
					__func__, __LINE__, *p);
				return -1;
			}
			i += 2;
			got_work++;
			//DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		case 'L':
			opts->db_list_owners = opt_yes;
			got_work++;
			DBG("%s:%d: found %c\n", __func__, __LINE__, *p);
			goto next_opt;
		default:
			fprintf(stderr, "unknown option: -%c\n\n", *p);
			return -1;
		}
	}
	//DBG(" <- %s:%d\n", __func__, __LINE__);
	return got_work ? 0 : -1;
}

static int process_op_list(const struct db_op_entry* db_op_list,
	struct os_area_db *db, const struct os_area_header *h, FILE *dev)
{
	int result = 0;
	const struct db_op_entry* e;
	int need_write = 0;

	for (e = db_op_list; e; e = e->next) {
		DBG("%s:%d: %p: type %d, (%d:%d) = %llu\n", __func__, __LINE__,
			e, e->type, e->id.owner, e->id.key,
			(unsigned long long)e->value);

		switch(e->type) {
		case op_print:
			os_area_db_print(db, &e->id);
			break;
		case op_remove:
			result = os_area_db_remove(db, &e->id);
			need_write = !result;
			break;
		case op_set_64:
			result = os_area_db_set_64(db, &e->id, e->value);
			need_write = !result;
			break;
		case op_set_32:
			result = os_area_db_set_32(db, &e->id, e->value);
			need_write = !result;
			break;
		case op_set_16:
			result = os_area_db_set_16(db, &e->id, e->value);
			need_write = !result;
			break;
		default:
			DBG("%s:%d: unknown type %d\n", __func__, __LINE__,
				e->type);
			result = -1;
			break;
		}

		if (result) {
			fprintf(stderr, "%s:%d: process_op_list failed\n",
				__func__, __LINE__);
			return -1;
		}
	}

	if (need_write) {
		os_area_db_write(db, h, dev);
		result = os_area_db_read(db, h, dev);
		
		if (result)
			fprintf(stderr, "%s:%d: process_op_list failed\n",
				__func__, __LINE__);
	}

	return result;
}

int main(int argc, char *argv[])
{
	int result;
	int found_db;
	struct opts opts;
	struct os_area_header header;
	struct os_area_params params;
	struct os_area_db db;
	FILE *dev;

	result = opts_parse(&opts, argc, argv);

	if (opts.verbosity >= 3)
		opts_dump(&opts);

	if (result) {
		print_usage();
		exit(1);
	}

	if (opts.show_help) {
		print_usage();
		exit(0);
	}

	if (opts.db_list_owners)
		db_list_owners();

	opts.device = opts.device ? opts.device : flash_dev;
	dev = fopen(opts.device, "r+");

	if (!dev) {
		DBG("%s:%d: fopen failed.\n", __func__, __LINE__);
		perror(opts.device);
		exit(1);
	}

	result = os_area_fixed_read(&header, &params, dev);

	if (result) {
		fprintf(stderr, "%s:%d: os_area_read_hp error.\n", __func__,
			__LINE__);
		exit(1);
	}

	if (opts.db_format) {
		result = os_area_db_format(&db, &header, dev);
		if (result) {
			fprintf(stderr, "%s:%d: db_format failed.\n",
				__func__, __LINE__);
			exit(1);
		}
	}

	found_db = !os_area_db_read(&db, &header, dev);
	
	if (opts.show_settings)
		show_settings(&header, &params, found_db ? &db : NULL);

	if (opts.image) {
		result = write_image(dev, &header, opts.image);
		if (result) {
			fprintf(stderr, "%s:%d: write_image failed: '%s'\n",
				__func__, __LINE__, opts.image);
			exit(1);
		}
	}

	if (opts.compressed_image) {
		result = os_area_set_ldr_format(&header, dev,
			opts.compressed_image == opt_yes);
		if (result) {
			fprintf(stderr, "%s:%d: set_compress_flag failed.\n",
				__func__, __LINE__);
			exit(1);
		}
	}

	if (opts.boot_other_os) {
		result = os_area_set_boot_flag(&params, dev,
			opts.boot_other_os == opt_yes);
		if (result) {
			fprintf(stderr, "%s:%d: set_boot_flag failed.\n",
				__func__, __LINE__);
			exit(1);
		}
	}

	if (opts.db_op_list) {
		result = process_op_list(opts.db_op_list, &db, &header, dev);

		if (result) {
			fprintf(stderr, "%s:%d: process_op_list failed.\n",
				__func__, __LINE__);
			exit(1);
		}
	}
	return 0;
}
